#!/usr/bin/env python
# -*- coding: utf-8 -*-
#

from hakkero.dataset.dataloader import PadLoader
from hakkero.dataset.dataloader import PreferencePadLoader
from hakkero.dataset.dataloader import PreferenceUnpadLoader
from hakkero.dataset.dataloader import UnpadLoader
from hakkero.dataset.image import process_messages
from hakkero.dataset.mixed_dataset import get_dataset
from hakkero.dataset.utils import IGNORE_INDEX



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#


from hakkero.dataset.strategy.errors import SegmentationError
from hakkero.dataset.strategy.errors import TokenizationError
from hakkero.dataset.strategy.segmentation import concat
from hakkero.dataset.strategy.segmentation import integrous
from hakkero.dataset.strategy.segmentation import naive
from hakkero.dataset.strategy.segmentation import unbiased
from hakkero.dataset.strategy.tokenization import chatml_message
from hakkero.dataset.strategy.tokenization import chatml_preference
from hakkero.dataset.strategy.tokenization import chatml_qwen2_vl_message
from hakkero.dataset.strategy.tokenization import huggingface_message
from hakkero.dataset.strategy.tokenization import huggingface_preference
from hakkero.dataset.strategy.tokenization import legacy

ST_INTEGROUS = "integrous"
ST_CONCAT = "concat"
ST_NAIVE = "naive"
ST_UNBIASED = "unbiased"

segment = {
    ST_INTEGROUS: integrous,
    ST_CONCAT: concat,
    ST_NAIVE: naive,
    ST_UNBIASED: unbiased,
}


ST_LEGACY = "legacy"
ST_HG = "hg"
ST_HG_PREFERENCE = "hg_preference"
ST_CHATML = "chatml"
ST_CHATML_PREFERENCE = "chatml_preference"
ST_CHATML_QWEN_VL = "chatml_qwen2_vl_message"

tokenize = {
    ST_LEGACY: legacy,
    ST_HG: huggingface_message,
    ST_HG_PREFERENCE: huggingface_preference,
    ST_CHATML: chatml_message,
    ST_CHATML_PREFERENCE: chatml_preference,
    ST_CHATML_QWEN_VL: chatml_qwen2_vl_message,
}


ST_SEGMENT = "st_segment"
ST_TOKENIZE = "st_tokenize"
default_strategy = {ST_SEGMENT: ST_NAIVE, ST_TOKENIZE: ST_LEGACY}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright @2025 AI. Inspur Inc.
#
# @author: sunxian <sunxian@inspur.com>
# @date: 2025/01/21
#


def _shard_partition(items, rank, world_size):
    src = (rank * len(items)) // world_size
    dst = ((rank + 1) * len(items)) // world_size
    return items[src:dst]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#

import itertools
import math
import os.path
import queue
import random
import threading
import time
from collections import defaultdict

import torch
import torch.utils.data
from bitarray import bitarray
from bitarray.util import deserialize
from bitarray.util import serialize
from tabulate import tabulate

from hakkero.dataset.indexed_dataset import IndexedDataset
from hakkero.dataset.logger import logger
from hakkero.dataset.utils import Range


class CacheEmpty(Exception):
    pass


class CacheFull(Exception):
    pass


class BlockedBitMap:
    def __init__(self, size, bsize):
        assert bsize & (bsize - 1) == 0

        self.size = size
        self.bsize = bsize
        self.last_bid = int(math.ceil(size / bsize)) - 1
        self.last_bsize = size % bsize if size % bsize else bsize
        self.obsolete = bitarray(int(math.ceil(size / bsize)))
        self.active = dict()

    def __repr__(self):
        s = f"obsolete({self.obsolete.count()}/{len(self.obsolete)}): {self.obsolete.to01()}"
        active = tabulate([[key, value.to01()] for key, value in self.active.items()], ["id", "bitarray"], "plain")

        if self.active:
            s += f"\nactive({len(self.active)}):\n{active}"

        return s

    def state(self):
        return serialize(self.obsolete), {k: serialize(v) for k, v in self.active.items()}

    @staticmethod
    def from_state(size, bsize, state):
        bitmap = BlockedBitMap(size, bsize)
        bitmap.obsolete = deserialize(state[0])
        bitmap.active = {k: deserialize(v) for k, v in state[1].items()}
        return bitmap

    def resize(self, bsize):
        if bsize == self.bsize:
            return

        assert bsize & (bsize - 1) == 0

        last_bid = int(math.ceil(self.size / bsize)) - 1
        last_bsize = self.size % bsize if self.size % bsize else bsize
        obsolete = bitarray(int(math.ceil(self.size / bsize)))
        active = dict()

        if self.bsize > bsize:
            ratio = self.bsize // bsize
            # expand obsolete bitarray
            for old_bid, used in enumerate(self.obsolete):
                if not used:
                    continue

                bid_st = old_bid * ratio
                bid_ed = min(bid_st + ratio, len(obsolete))
                obsolete[bid_st:bid_ed] = 1
            # split active bitarray and mark exhausted ones in obsolete
            for old_bid, abarray in self.active.items():
                for i, sid_st in enumerate(range(0, len(abarray), bsize)):
                    bid = old_bid * ratio + i
                    sid_ed = min(sid_st + bsize, len(abarray))
                    if abarray[sid_st:sid_ed].all():
                        obsolete[bid] = 1
                    else:
                        active[bid] = abarray[sid_st:sid_ed]
        else:
            ratio = bsize // self.bsize
            # concat active bitarrays
            for old_bid, abarray in self.active.items():
                bid = old_bid // ratio
                if bid not in active:
                    if bid == last_bid:
                        active[bid] = bitarray(last_bsize)
                    else:
                        active[bid] = bitarray(bsize)

                sid_st = old_bid % ratio * self.bsize
                sid_ed = min(sid_st + self.bsize, len(active[bid]))
                active[bid][sid_st:sid_ed] |= abarray

            # shrink obsolete bitarray, update active bitarrays
            for bid, old_bid_st in enumerate(range(0, len(self.obsolete), ratio)):
                old_bid_ed = min(old_bid_st + ratio, len(self.obsolete))
                if self.obsolete[old_bid_st:old_bid_ed].all():
                    # the merged block is obsolete
                    obsolete[bid] = 1
                else:
                    # the merged block is active, mark the used samples
                    if bid not in active:
                        if bid == last_bid:
                            active[bid] = bitarray(last_bsize)
                        else:
                            active[bid] = bitarray(bsize)

                    for i, used in enumerate(self.obsolete[old_bid_st:old_bid_ed]):
                        if not used:
                            continue

                        sid_st = i * self.bsize
                        sid_ed = min(sid_st + self.bsize, len(active[bid]))
                        active[bid][sid_st:sid_ed] = 1

        for bid in list(active.keys()):
            if not active[bid].any():
                del active[bid]

        self.bsize = bsize
        self.last_bid = last_bid
        self.last_bsize = last_bsize
        self.obsolete = obsolete
        self.active = active

    def set(self, indices):
        group = defaultdict(list)
        for i in indices:
            group[i // self.bsize].append(i % self.bsize)

        for bid, sids in group.items():
            if self.obsolete[bid]:
                continue

            if bid not in self.active:
                if bid == self.last_bid:
                    self.active[bid] = bitarray(self.last_bsize)
                else:
                    self.active[bid] = bitarray(self.bsize)

            try:
                self.active[bid][sids] = 1
            except Exception as e:
                raise e

            if self.active[bid].all():
                self.obsolete[bid] = 1
                del self.active[bid]

    def __eq__(self, other):
        if self.obsolete != other.obsolete:
            return False

        if self.active.keys() != other.active.keys():
            return False

        for bid, abarray in self.active.items():
            if other.active[bid] != abarray:
                return False

        return True

    def __iadd__(self, other):
        assert self.size == other.size
        assert self.bsize == other.bsize

        self.obsolete |= other.obsolete
        for bid, abarray in other.active.items():
            if bid in self.active:
                self.active[bid] |= abarray
                if self.active[bid].all():
                    self.obsolete[bid] = 1
                    del self.active[bid]
            else:
                self.active[bid] = abarray

        return self

    def block_used(self, bid):
        if self.obsolete[bid]:
            return True

    def sample_used(self, index):
        bid = index // self.bsize
        if self.obsolete[bid]:
            return True
        elif index is not None and bid in self.active and self.active[bid][index % self.bsize]:
            return True
        else:
            return False

    def all(self):
        return self.obsolete.all()


class IteratorState:
    def __init__(self, size, bsize):
        self.size = size
        self.bsize = bsize
        self.epoch = 0
        self.bitmaps = dict()  # {epoch: BlockedBitMap}

    def __repr__(self):
        s = f"IteratorState({self.size}, {self.bsize}) @ epoch {self.epoch}"
        return "\n".join([s] + [f"epoch {epoch}:\n{repr(bitmap)}" for epoch, bitmap in self.bitmaps.items()])

    def block_used(self, epoch, bid):
        if epoch < self.epoch:
            return True
        if epoch not in self.bitmaps:
            return False

        return self.bitmaps[epoch].block_used(bid)

    def sample_used(self, epoch, index):
        if epoch < self.epoch:
            return True
        if epoch not in self.bitmaps:
            return False

        return self.bitmaps[epoch].sample_used(index)

    def track(self, stats):
        # stats: {epoch: indices}
        for epoch, indices in stats.items():
            if epoch < self.epoch:
                continue
            if epoch not in self.bitmaps:
                self.bitmaps[epoch] = BlockedBitMap(self.size, self.bsize)
            self.bitmaps[epoch].set(indices)

            if self.bitmaps[epoch].all():
                self.epoch = epoch + 1
                del self.bitmaps[epoch]

    def state_dict(self):
        return {
            "epoch": self.epoch,
            "size": self.size,
            "bsize": self.bsize,
            "bitmaps": {epoch: bitmap.state() for epoch, bitmap in self.bitmaps.items()},
        }

    def cleanup(self):
        # clean up exhausted bitmaps and update self.epoch
        epochs = sorted(self.bitmaps.keys(), reverse=True)
        for epoch in epochs:
            if self.bitmaps[epoch].all():
                # epoch start from the largest exhausted epoch
                self.epoch = epoch + 1
                break

        for epoch in epochs:
            # clean up exhausted bitmaps to save memory
            if epoch < self.epoch:
                del self.bitmaps[epoch]

    @staticmethod
    def from_state_dict(state_dict):
        # wrap_data = True: directly wrap the state_dict instead of copying to save space
        state = IteratorState(state_dict["size"], state_dict["bsize"])
        state.epoch = state_dict["epoch"]
        state.bitmaps = {
            k: BlockedBitMap.from_state(state.size, state.bsize, v) for k, v in state_dict["bitmaps"].items()
        }
        return state

    def reset(self):
        self.epoch = 0
        self.bitmaps = dict()

    def __iadd__(self, other):
        if self.bsize != other.bsize:
            other.resize(self.bsize)

        for epoch, bitmap in other.bitmaps.items():
            if epoch in self.bitmaps:
                self.bitmaps[epoch] += bitmap
            else:
                self.bitmaps[epoch] = bitmap

        return self

    def update(self, state_dict):
        state = IteratorState.from_state_dict(state_dict)
        self.__iadd__(state)


class IterableDataset(IndexedDataset, torch.utils.data.IterableDataset):
    def __init__(
        self,
        path,
        name=None,
        seed=-1,
        max_epoch=1,
        block_size=1024,
        prefetcher=None,
        infinite=False,
        n_shards=1,
        rank=0,
        world_size=1,
    ):
        super().__init__(path)

        self.name = name if name else os.path.basename(path)
        self.rank, self.world_size = rank, world_size
        self.n_shards = n_shards

        # infinite=False: stop immediately when current epoch exceeds self.max_epoch
        # infinite=True: simple mark self.exhausted and keep iterating
        self.infinite = infinite
        self.max_epoch = max_epoch
        self.seed = seed
        self.block_size, self.n_blocks = self.safe_block(block_size)

        self.lock = threading.RLock()
        self.cache = self.blocks = None  # lock protected variables
        self.prefetcher = prefetcher
        self.state = IteratorState(len(self), self.block_size)

        # intermediate states
        self.block_iter = None
        self.random = None
        self.exhausted = None

    def load_state_dict(self, state_or_state_dict):
        if isinstance(state_or_state_dict, IteratorState):
            self.state = state_or_state_dict
        else:
            self.state = IteratorState.from_state_dict(state_or_state_dict)

    def state_dict(self):
        return self.state.state_dict()

    def track(self, stats):
        self.state.track(stats)

    def safe_block(self, block_size):
        assert (block_size & (block_size - 1)) == 0, f"block_size should be power of 2, but got {block_size}."

        if len(self) < self.n_shards:
            raise ValueError(f"cannot split {len(self)} entries of {self.path} into {self.n_shards} shards")

        n_blocks = int(math.ceil(len(self) / block_size))
        if n_blocks < self.n_shards:
            block_size = len(self) // self.n_shards
            block_size = int(math.pow(2, math.floor(math.log(block_size) / math.log(2))))
            n_blocks = int(math.ceil(len(self) / block_size))

        return block_size, n_blocks

    @property
    def priority(self):
        return 2 - len(self.cache)

    def prefetch(self):
        with self.lock:
            if len(self.cache) == 2:
                raise CacheFull

            block = next(self.blocks)
            st = block[1] * self.block_size
            ed = min(st + self.block_size, len(self))

            bytes = self.get_bytes(slice(st, ed))
            samples = [dict(info=(block[0], i), byte=byte) for i, byte in zip(range(st, ed), bytes)]
            self.cache.append(samples)

    def build_next_iter(self, samples):
        if self.random is not None:
            self.random.shuffle(samples)

        for sample in samples:
            epoch, index = sample["info"]
            if self.state.sample_used(epoch, index):
                continue

            data = self.decode_byte(index, sample.pop("byte"))
            if data:
                sample["data"] = data["data"]
                sample["uid"] = data["uid"]

            yield sample

    def next(self, block=False):
        # handle transition across blocks
        # raise CacheNotFilled when cache is to be filled
        # when infinite = False: raise StopIteration when the dataset is exhausted
        # when infinite = True: silently mark exhausted = True
        while True:
            if self.exhausted and not self.infinite:
                raise StopIteration(f"Exhausted dataset '{self.path}'")

            try:
                return next(self.block_iter)
            except StopIteration:
                while not self.cache:
                    if block:
                        time.sleep(0.1)  # wait for the prefetcher to load next block
                    else:
                        raise CacheEmpty  # signal prefetcher to load next block

                samples = self.cache.pop(0)
                epoch = samples[0]["info"][0]
                if epoch >= self.max_epoch:  # epoch start from 0
                    self.exhausted = True

                self.block_iter = self.build_next_iter(samples)

    def __next__(self):
        if self.prefetcher is None:
            self.prefetcher = Prefetcher()

        return self.prefetcher.next(self)

    def __iter__(self):
        worker_info = torch.utils.data.get_worker_info()
        if worker_info is None:
            n_workers, worker_id = 1, 0
        else:
            n_workers, worker_id = worker_info.num_workers, worker_info.id

        r = Range(0, self.n_blocks, 1)
        r = r.sub_range(split=self.rank, n_splits=self.world_size)  # split index among multi-gpu workers
        r = r.sub_range(split=worker_id, n_splits=n_workers)  # split index among multiprocess dataloader workers

        logger.debug(
            f"world_size: {self.world_size}, rank: {self.rank}, n_workers: {n_workers}, worker_id: {worker_id}, r={r.list()}"
        )

        if self.seed >= 0:
            blocks = ((epoch, bid) for epoch in itertools.count(self.state.epoch) for bid in r.random_iterate())
            self.random = random.Random(self.seed)
        else:
            blocks = ((epoch, bid) for epoch in itertools.count(self.state.epoch) for bid in r.iterate())
            self.random = None

        blocks = itertools.filterfalse(lambda block: self.state.block_used(block[0], block[1]), blocks)

        self.exhausted = self.state.epoch >= self.max_epoch

        self.block_iter = iter(())
        with self.lock:
            self.cache = []
            self.blocks = blocks

        return self


class Prefetcher:
    # use a joint thread pool to prefetch multiple datasets random access of a list of PrefetchIterableDataset
    # prioritize prefetching dataset that is being consumed
    def __init__(self):
        self.datasets = []
        self.urgent = queue.Queue()
        self.should_stop = threading.Event()
        self.thread = None

    def stop_thread(self):
        if self.alive:
            self.should_stop.set()
            self.thread.join(timeout=60)

    @property
    def alive(self):
        return isinstance(self.thread, threading.Thread) and self.thread.is_alive()

    def __del__(self):
        self.stop_thread()

    def worker(self):
        # prefetch blocks of each dataset
        while not self.should_stop.is_set():
            try:
                dataset = self.urgent.get_nowait()  # prioritize urgent load
            except queue.Empty:
                if self.datasets:
                    # prioritize the least cached datasets
                    dataset = max(self.datasets, key=lambda d: d.priority)
                else:
                    time.sleep(0.001)
                    continue

            try:
                dataset.prefetch()
            except (CacheFull, StopIteration):
                time.sleep(0.001)

    def next(self, dataset):
        if not self.alive:
            self.thread = threading.Thread(target=self.worker, daemon=True)
            self.thread.start()

        if dataset not in self.datasets:
            self.datasets.append(dataset)

        try:
            return dataset.next(block=False)
        except CacheEmpty:
            self.urgent.put(dataset)
            return dataset.next(block=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#


LEGACY_KEYS = ["title", "summary", "abstract", "text", "question", "answer", "code"]
FMT_MESSAGES = "[{'role': 'x', 'content': 'x'}, ...]"
FMT_PREFERENCES = "{'context': [{'role': 'x', 'content': 'x'}, ...], 'chosen': 'x', 'rejected': 'x'}"


def check_legacy(data):
    if not isinstance(data, dict):
        return False, "wrong data format, expect {key: value}, " + f"but got {data}"

    if all(s not in data for s in LEGACY_KEYS):
        return False, f"No valid keys in data, expect of: ({LEGACY_KEYS})"

    if sum([len(data[key]) for key in LEGACY_KEYS]) == 0:
        return False, "all valid keys in data are empty"

    return True, ""


def check_message(data):
    if len(data) == 0:
        return False, "messages should not be empty"

    if not isinstance(data, list) or not all(isinstance(d, dict) for d in data):
        return False, f"messages should be {FMT_MESSAGES}, but got {data}"

    if data[0]["role"] not in ("system", "user"):
        return False, f"messages[0]['role'] should be 'system' or 'user', but got {data[0]['role']}"

    if data[-1]["role"] != "assistant":
        return False, f"messages[-1]['role'] should be 'assistant', but got {data[-1]['role']}"

    if data[-2]["role"] != "user":
        return False, "messages[-2]['role'] should be 'user'"

    if data[0]["role"] == "system" and len(data[1:]) % 2 != 0:
        return False, f"messages should be in pairs between user and assistant, but got {data}"

    if any([len(d["content"]) == 0 for d in data]):
        return False, f"messages should not be empty, but some of them are empty. see {data}"

    return True, ""


def check_preference(data):
    if not isinstance(data, dict):
        return False, f"messages should be {FMT_PREFERENCES}, but got {data}"

    if "context" not in data or "chosen" not in data or "rejected" not in data:
        return False, f"messages should be {FMT_PREFERENCES}, but got {data}"

    if not isinstance(data["chosen"], str) or not isinstance(data["rejected"], str):
        return False, f"chosen or rejected should be string, but got {data}"

    # check context
    context = data["context"]
    if len(context) == 0:
        return False, f"context should not be empty, but got {data}"

    if not isinstance(context, list) or not all(isinstance(d, dict) for d in context):
        return False, f"context should be {FMT_MESSAGES}, but got {data}"

    if context[0]["role"] not in ("system", "user"):
        return False, f"context[0]['role'] should be 'system' or 'user', but got {context[0]['role']}"

    if context[-1]["role"] != "user":
        return False, f"context[-1]['role'] should be 'user', but got {context[-1]['role']}"

    if context[0]["role"] == "system" and (len(context[1:]) + 1) % 2 != 0:
        return False, f"context should be in pairs between user and assistant, but got {data}"

    if any([len(d["content"]) == 0 for d in context]):
        return False, f"message in context should not be empty, but some of them are empty. see {data}"

    return True, ""


check_func = {
    "legacy": check_legacy,
    "message": check_message,
    "preference": check_preference,
}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#

import argparse
import os
import sys
from subprocess import check_output

ROOT_PATH = os.path.dirname(__file__)
if getattr(sys, "frozen", False):
    ROOT_PATH = os.path.dirname(sys.executable)

_turboshuf = os.path.join(ROOT_PATH, "turboshuf")


def turboshuf(filename, output, seed=None, skip=0, memory=4.0):
    if not os.path.isfile(filename):
        raise FileNotFoundError(f"file not found: {filename}")

    full = os.path.abspath(filename)
    parent_path = os.path.dirname(full)
    name, ext = os.path.splitext(os.path.basename(full))

    tmp_dir = os.path.join(parent_path, f"{name}-tmp")
    os.makedirs(tmp_dir, exist_ok=True)

    env = {
        "SKIP": skip,
        "TMPDIR": tmp_dir,
        "MEMORY": memory,
        "SEED": seed,
    }
    envs = " ".join([f"{k}={v}" for k, v in env.items() if v is not None])

    cmd = f"{envs} {_turboshuf} <{filename} >{output}"
    try:
        check_output(cmd, shell=True, text=True)
    finally:
        os.rmdir(tmp_dir)

    return True


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="shuf data fast")
    parser.add_argument("--filename", type=str)
    parser.add_argument("--output", type=str)
    parser.add_argument("--seed", type=int, default=None)
    parser.add_argument("--skip", type=int, default=0)
    parser.add_argument("--memory", type=float, default=4.0)

    args = parser.parse_args()

    turboshuf(filename=args.filename, output=args.output, seed=args.seed, skip=args.skip, memory=args.memory)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#

import numpy as np
import torch.utils.data
from tabulate import tabulate

from hakkero.dataset.iterable_dataset import Prefetcher
from hakkero.dataset.segment_dataset import MixedSegmentDataset
from hakkero.dataset.strategy import default_strategy
from hakkero.dataset.strategy import ST_SEGMENT
from hakkero.dataset.strategy import ST_TOKENIZE
from hakkero.dataset.utils import MultinomialSampler


class DummyDataset(torch.utils.data.Dataset):
    def __getitem__(self, item):
        return 0

    def __len__(self):
        return 100000


class MixedDataset(torch.utils.data.IterableDataset):
    def __init__(
        self,
        config,
        tokenizer,
        num_epochs=1,
        max_length=1024,
        homogeneous=False,
        seed=-1,
        rank=0,
        world_size=1,
        n_shards=2,
        **kwargs,
    ):
        super().__init__()

        self.tokenizer = tokenizer
        self.num_epochs = num_epochs
        self.max_length = max_length
        self.homogeneous = homogeneous
        self.norm_weight_with_n_targets = kwargs.pop("norm_weight_with_n_targets", False)

        self.seed = seed

        self.rank = rank
        self.world_size = world_size
        self.n_shards = n_shards

        self.datasets = []
        self.weights = []
        self.prefetcher = Prefetcher()

        strategy_segment = kwargs.pop(ST_SEGMENT, None) or default_strategy[ST_SEGMENT]
        strategy_tokenize = kwargs.pop(ST_TOKENIZE, None) or default_strategy[ST_TOKENIZE]

        for i, (name, sub_config) in enumerate(config.items()):
            if isinstance(sub_config, str):
                path = sub_config
                n_epoch = 1 * self.num_epochs if self.num_epochs > 0 else 1
                strategy = {ST_SEGMENT: strategy_segment, ST_TOKENIZE: strategy_tokenize}
                weight = None
            else:
                path = sub_config["path"]
                sub_config_epoch = int(sub_config.get("epoch", 1))
                n_epoch = sub_config_epoch * self.num_epochs if self.num_epochs > 0 else sub_config_epoch
                strategy = sub_config.get("strategy", {})

                if ST_SEGMENT not in strategy:
                    strategy[ST_SEGMENT] = strategy_segment

                if ST_TOKENIZE not in strategy:
                    strategy[ST_TOKENIZE] = strategy_tokenize

                weight = sub_config.get("weight", None)

            dataset = MixedSegmentDataset(
                path,
                name=name,
                tokenizer=tokenizer,
                prefetcher=self.prefetcher,
                seed=self.seed,
                infinite=True if homogeneous else False,
                max_epoch=n_epoch,
                strategy=strategy,
                max_length=max_length,
                n_shards=n_shards,
                rank=rank,
                world_size=world_size,
                **kwargs,
            )

            self.datasets.append(dataset)
            # weight by size by default, can be overwritten with load_weights
            self.weights.append(n_epoch * dataset.size if weight is None else weight)

        self.weights = np.array(self.weights, dtype=np.float64)
        self.weights /= self.weights.sum()

        self.sampler = MultinomialSampler(seed)
        self.active = np.ones_like(self.weights)

    def __iter__(self):
        self.sampler = MultinomialSampler(self.seed)
        self.active = np.ones_like(self.weights)
        for d in self.datasets:
            iter(d)

        while True:
            if all(d.exhausted for d in self.datasets):
                return

            weights = self.active * self.weights
            if self.norm_weight_with_n_targets:
                # remap the desired token distribution to sample distribution
                # as different dataset has different average #tokens
                n_targets = np.array([d.avg_n_target for d in self.datasets], dtype=np.float64)
                weights = weights / n_targets

            try:
                i = self.sampler.next(weights)
            except StopIteration:
                return

            try:
                segment = next(self.datasets[i])
                segment["task"] = i
                yield segment
            except StopIteration:
                self.active[i] = 0
                continue

    def state_dict(self):
        return {
            "datasets": {d.name: d.state_dict() for d in self.datasets},
            "weights": {d.name: w for d, w in zip(self.datasets, self.weights)},
        }

    def load_state_dict(self, state_dict):
        names = [d.name for d in self.datasets]
        self.weights = [state_dict["weights"].get(name, 1.0) for name in names]
        assert all(w >= 0.0 for w in self.weights), "loaded weights should be greater than 0"
        for dataset in self.datasets:
            if dataset.name not in state_dict["datasets"]:
                continue
            dataset.load_state_dict(state_dict["datasets"][dataset.name])
        missing_keys = list(set(names) - set(state_dict["datasets"].keys()))

        return missing_keys

    def load_weights(self, weight_dict):
        names = [d.name for d in self.datasets]
        missing_keys = list(set(names) - set(weight_dict.keys()))
        if missing_keys:
            raise ValueError(f"missing keys of weight_dict: {missing_keys}")
        self.weights = [weight_dict.get(name, 1.0) for name in names]
        assert all(w >= 0.0 for w in self.weights), "loaded weights should be greater than 0"

    def track(self, task_stats):
        # task_stats = {task: {epoch: [indices]}}
        for task, stats in task_stats.items():
            self.datasets[task].track(stats)

    def __repr__(self):
        entries = (
            (d.name, w, f"{len(d):,}", d.size / 1024**3, d.size / len(d) / 1024)
            for w, d in zip(self.weights, self.datasets)
        )
        header = ["name", "weight", "#entries", "size(GB)", "average len(KB)"]
        table = tabulate(
            sorted(entries, key=lambda x: x[1], reverse=True),
            header,
            "pipe",
            floatfmt=".6f",
            colalign=("left", "right", "right", "right"),
        )
        return f"{self.__class__.__name__} with \n{table}"


def get_dataset(
    config,
    tokenizer,
    num_epochs,
    max_length=4096,
    homogeneous=False,
    seed=-1,
    rank=0,
    world_size=1,
    n_workers=2,
    **kwargs,
):
    return MixedDataset(
        config,
        tokenizer,
        num_epochs=num_epochs,
        max_length=max_length,
        homogeneous=homogeneous,
        seed=seed,
        rank=rank,
        world_size=world_size,
        n_shards=n_workers * world_size,
        **kwargs,
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#

import os
from copy import deepcopy
from io import BytesIO
from typing import List
from typing import Optional
from typing import TypedDict
from typing import Union

from PIL import Image
from PIL.Image import Image as ImageObject
from PIL.Image import Resampling


class EncodedImage(TypedDict):
    path: Optional[str]
    bytes: Optional[bytes]


ImageInput = Union[str, EncodedImage, ImageObject]
IMAGE_PLACEHOLDER = "<image>"


def preprocess(image: "ImageObject", **kwargs) -> "ImageObject":
    image_resolution: int = kwargs.get("image_resolution")

    if max(image.width, image.height) > image_resolution:
        resize_factor = image_resolution / max(image.width, image.height)
        width, height = int(image.width * resize_factor), int(image.height * resize_factor)
        image = image.resize((width, height), resample=Resampling.NEAREST)

    if image.mode != "RGB":
        image = image.convert("RGB")
    if min(image.width, image.height) < 28:
        width, height = max(image.width, 28), max(image.height, 28)
        image = image.resize((width, height), resample=Resampling.NEAREST)

    if image.width / image.height > 200:
        width, height = image.height * 180, image.height
        image = image.resize((width, height), resample=Resampling.NEAREST)

    if image.height / image.width > 200:
        width, height = image.width, image.width * 180
        image = image.resize((width, height), resample=Resampling.NEAREST)

    return image


def regularize(images, **kwargs) -> List["ImageObject"]:
    r"""
    Regularizes images to avoid error. Including reading and pre-processing.
    """
    results = []
    for image in images:
        if isinstance(image, str):
            image = Image.open(image)
        elif isinstance(image, dict):
            if image["bytes"] is not None:
                image = Image.open(BytesIO(image["bytes"]))
            else:
                image = Image.open(image["path"])

        if not isinstance(image, ImageObject):
            raise ValueError("Expect input is a list of Images, but got {}.".format(type(image)))

        results.append(preprocess(image, **kwargs))

    return results


def _get_mm_inputs(images, processor):
    image_processor = getattr(processor, "image_processor")
    video_processor = getattr(processor, "video_processor", image_processor)
    input_dict = {"images": None}
    if len(images) != 0:
        images = regularize(
            images,
            image_resolution=getattr(processor, "image_resolution", 512),
        )
        input_dict["images"] = images

    mm_inputs = {}
    if input_dict.get("images") is not None:
        mm_inputs.update(image_processor(input_dict["images"], return_tensors="pt"))

    if input_dict.get("videos") is not None:
        mm_inputs.update(video_processor(input_dict["videos"], return_tensors="pt"))

    return mm_inputs


def translate_messages(messages, mm_path):
    res_messages = []
    images = []
    for i, message in enumerate(messages):
        role = message["role"]

        txt = ""
        for content in message["content"]:
            if content["type"] == "image":
                images.append(os.path.join(mm_path, content["image"]))
                txt += "<image>"
            elif content["type"] == "text":
                txt += content["text"]
            else:
                raise NotImplementedError("not supported type: {}".format(content["type"]))

        res_messages.append({"role": role, "content": txt})
    return res_messages, images


def process_messages(messages, images, processor, image_pad_token="<|image_pad|>"):
    image_processor = getattr(processor, "image_processor")
    merge_length = getattr(image_processor, "merge_size") ** 2
    mm_inputs = _get_mm_inputs(images, processor)
    image_grid_thw = mm_inputs.get("image_grid_thw", [])

    num_image_tokens = 0
    messages = deepcopy(messages)
    for message in messages:
        content = message["content"]
        while IMAGE_PLACEHOLDER in content:
            if num_image_tokens >= len(image_grid_thw):
                raise ValueError("`len(images)` is less than the number of {} tokens.".format(IMAGE_PLACEHOLDER))

            content = content.replace(
                IMAGE_PLACEHOLDER,
                "<|vision_start|>{}<|vision_end|>".format(
                    image_pad_token * (image_grid_thw[num_image_tokens].prod() // merge_length)
                ),
                1,
            )
            num_image_tokens += 1

        message["content"] = content

    if len(images) != num_image_tokens:
        raise ValueError("The number of images does not match the number of {} tokens".format(IMAGE_PLACEHOLDER))

    return messages, mm_inputs



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#

import os.path
import random

import numpy as np
import torch.utils.data

from hakkero.dataset.iterable_dataset import IterableDataset
from hakkero.dataset.logger import logger
from hakkero.dataset.strategy import default_strategy
from hakkero.dataset.strategy import segment as strategy_segment
from hakkero.dataset.strategy import SegmentationError
from hakkero.dataset.strategy import ST_CONCAT
from hakkero.dataset.strategy import ST_SEGMENT
from hakkero.dataset.strategy import ST_TOKENIZE
from hakkero.dataset.strategy import TokenizationError
from hakkero.dataset.strategy import tokenize as strategy_tokenize
from hakkero.dataset.utils import MultinomialSampler
from hakkero.dataset.utils import RunningAverage


class SegmentDataset(torch.utils.data.IterableDataset):
    def __init__(
        self,
        path,
        name=None,
        tokenizer=None,
        prefetcher=None,
        seed=-1,
        infinite=False,
        max_epoch=1,
        strategy=None,
        max_length=1024,
        n_shards=2,
        rank=0,
        world_size=1,
        **kwargs,
    ):
        super().__init__()

        self.path = path
        self.name = name if name else os.path.basename(path)
        self.seed = seed
        self.rank = rank
        self.world_size = world_size

        block_size = kwargs.pop("block_size", 1024)
        self.dataset = IterableDataset(
            path,
            name=name,
            max_epoch=max_epoch,
            n_shards=n_shards,
            seed=seed,
            block_size=block_size,
            prefetcher=prefetcher,
            infinite=infinite,
            rank=rank,
            world_size=world_size,
        )

        self.strategy = default_strategy if strategy is None else strategy

        self.tokenizer = tokenizer
        self.processor = kwargs.pop("processor", None)
        self.max_length = max_length

        self.prev = None
        self.length = RunningAverage()
        self.n_target = RunningAverage()
        self.random = None if self.seed < 0 else random.Random(self.seed)

        self.kwargs = kwargs

    def __next__(self):
        sample = next(self.dataset)
        if "data" not in sample:
            return [dict(used=[sample["info"]])]

        try:
            data = strategy_tokenize[self.strategy[ST_TOKENIZE]](
                sample["data"], self.tokenizer, processor=self.processor, path=self.path, **self.kwargs
            )
        except TokenizationError as e:
            logger.warning(
                f"[rank={self.rank}, world_size={self.world_size}][{self.path}:{sample['info'][1]}]: {e} => {sample['uid']}\n{sample['data']}"
            )
            return [dict(used=[sample["info"]])]

        try:
            if self.strategy[ST_SEGMENT] == ST_CONCAT:
                while True:
                    segments, self.prev = strategy_segment[self.strategy[ST_SEGMENT]](
                        data, self.max_length, sample["info"], self.random, self.prev
                    )
                    if segments:
                        break
            else:
                segments = strategy_segment[self.strategy[ST_SEGMENT]](
                    data, self.max_length, sample["info"], self.random
                )
        except SegmentationError as e:
            logger.warning(
                f"[rank={self.rank}, world_size={self.world_size}][{self.path}:{sample['info'][1]}]: {e} => {sample['uid']}"
            )
            return [dict(used=[sample["info"]])]

        for s in segments:
            if "labels" in s:
                # preference data
                assert "inputs" in s
                s["length"] = {key: value.nelement() for key, value in s["labels"].items()}
                s["n_target"] = {key: value.gt(0).count_nonzero().item() for key, value in s["labels"].items()}
                self.length.update(list(s["length"].values()))
                self.n_target.update(list(s["n_target"].values()))
            else:
                # pretrain or sft
                s["length"] = s["label"].nelement()
                s["n_target"] = s["label"].gt(0).count_nonzero().item()
                self.length.update(s["length"])
                self.n_target.update(s["n_target"])

        return segments

    def __iter__(self):
        iter(self.dataset)
        self.prev = None
        self.length = RunningAverage()
        self.n_target = RunningAverage()
        self.random = None if self.seed < 0 else random.Random(self.seed)
        return self

    @property
    def avg_length(self):
        assert self.length.value > 0, f"{self.path} has zero length, please ensure the dataset is well-formed"
        return self.length.value

    @property
    def avg_n_target(self):
        assert self.n_target.value > 0, f"{self.path} has zero target tokens, please ensure the dataset is well-formed"
        return self.n_target.value

    def state_dict(self):
        # {"size": size, "epoch": epoch, "bsize": block_size, "states": {epoch: BlockedBitMap()}}
        return self.dataset.state_dict()

    def load_state_dict(self, state_or_state_dict):
        self.dataset.load_state_dict(state_or_state_dict)

    def track(self, stats):
        # states: {epoch: indices}
        self.dataset.track(stats)


class MixedSegmentDataset(SegmentDataset):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # fewer slots for less mixed distribution but faster initial loading
        self.n_slots = 1 if self.seed < 0 else 20
        self.weights = np.array([1 / self.n_slots] * self.n_slots)
        self.sampler = MultinomialSampler(self.seed)
        self.slots = [[] for _ in range(self.n_slots)]
        self.remain = None
        self.exhausted = False

    @property
    def size(self):
        return self.dataset.size

    def __len__(self):
        return len(self.dataset)

    def __iter__(self):
        super().__iter__()
        self.weights = np.array([1 / self.n_slots] * self.n_slots, dtype=np.float64)
        self.sampler = MultinomialSampler(self.seed if self.seed >= 0 else 0)
        self.slots = [[] for _ in range(self.n_slots)]
        self.remain = None
        self.exhausted = False

        return self

    def __next__(self):
        while True:
            # select a non-empty slot
            try:
                i = self.sampler.next(self.weights)
            except StopIteration:
                self.exhausted = True
                raise StopIteration()

            if not self.slots[i]:
                try:
                    # try to fill the slot
                    self.slots[i] = super().__next__()
                except StopIteration:
                    # no new segments, mask the slot and move on
                    self.weights[i] = 0
                    continue

            # the current slot is selected and filled
            break

        # maybe start count down for the infinite case
        if self.dataset.exhausted:
            if self.remain is None:
                self.remain = sum(len(slot) for slot in self.slots)

            self.remain -= 1

            if self.remain <= 0:
                self.exhausted = True

        return self.slots[i].pop(0)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright @2025 AI. Inspur Inc.
#
# @author: sunxian <sunxian@inspur.com>
# @date: 2025/01/21
#
import torch.utils.data


def get_dummy_dataloader(max_len, batch_size, vocab_size, rank, world_size):
    class SteadyCounter(torch.utils.data.IterableDataset):
        def __init__(self, max_len, vocab_size):
            self.i = 0

            self.max_len = max_len
            self.vocab_size = vocab_size

        def __iter__(self):
            while True:
                out = torch.Tensor([x % self.vocab_size for x in range(self.i, self.i + self.max_len)]).long()
                yield out, out

                self.i += self.max_len

    data = SteadyCounter(max_len, vocab_size)
    return torch.utils.data.DataLoader(data, batch_size=batch_size)


def get_dataloader(cfg, rank, world_size):
    pass
